from typing import Optional

import numpy as np
from scipy.integrate import simpson

from .pattern import Pattern
from .methods import FourierTransformMethod


def calculate_sq(
    normalized_pattern: Pattern,
    f_squared_mean: np.ndarray,
    f_mean_squared: np.ndarray,
) -> Pattern:
    """
    Calculates the Faber Ziman structure factor, using the equation:
    S(Q) = (n * Intensity - incoherent_scattering - <f>^2-)/<f^2> + 1
    where n is the normalization factor and f are the scattering factors.
    The function takes in the already normalized intensity (incoherent scattering is also subtracted)

    :param normalized_pattern:    Pattern with q and (n * Intensity - incoherent scattering) as x and y
    :param f_squared_mean:        <f^2> - mean squared scattering factor for each q value in the pattern
    :param f_mean_squared:        <f>^2 - squared mean scattering factor for each q value in the pattern

    :return: S(Q) pattern
    """
    sq = (normalized_pattern.y - f_squared_mean + f_mean_squared) / f_mean_squared
    return Pattern(normalized_pattern.x, sq)


def calculate_fr(
    sq_pattern: Pattern,
    r: Optional[np.ndarray] = None,
    use_modification_fcn: bool = False,
    method: str = "integral",
) -> Pattern:
    """
    Calculates F(r) from a given S(Q) pattern for r values.
    If r is None, a range from 0 to 10 with step 0.01 is used.
    A Lorch modification function of the form:

        m = sin(q*pi/q_max)/(q*pi/q_max)

    can be used to address issues with a low q_max. This will broaden the sharp peaks in g(r)

    :param sq_pattern:              Structure factor S(Q) with lim_inf S(Q) = 1 and unit(q)=A^-1
    :param r:                       numpy array giving the r-values for which F(r) will be calculated,
                                    default is 0 to 10 with 0.01 as a step. units should be in Angstrom.
    :param use_modification_fcn:    boolean flag whether to use the Lorch modification function
    :param method:                  determines the method used for calculating fr, possible values are:
                                            - 'integral' solves the Fourier integral, by calculating the integral
                                            - 'fft' solves the Fourier integral by using fast fourier transformation

    :return: F(r) pattern
    """
    if r is None:
        r = np.linspace(0.01, 10, 1000)

    q, sq = sq_pattern.data
    if use_modification_fcn:
        modification = np.sin(q * np.pi / np.max(q)) / (q * np.pi / np.max(q))
    else:
        modification = 1

    if method == "integral" or method == FourierTransformMethod.INTEGRAL:
        fr = (
            2.0
            / np.pi
            * simpson(
                modification * q * (sq - 1) * np.array(np.sin(np.outer(q.T, r))).T, x=q
            )
        )
    elif method == "fft" or method == FourierTransformMethod.FFT:
        q_step = q[1] - q[0]
        r_step = r[1] - r[0]

        n_out = np.max([len(q), int(np.pi / (r_step * q_step))])
        q_max_for_ifft = 2 * n_out * q_step
        y_for_ifft = np.concatenate(
            (modification * q * (sq - 1), np.zeros(2 * n_out - len(q)))
        )

        ifft_result = np.fft.ifft(y_for_ifft) * 2 / np.pi * q_max_for_ifft
        ifft_imag = np.imag(ifft_result)[:n_out]
        ifft_x_step = 2 * np.pi / q_max_for_ifft
        ifft_x = np.arange(n_out) * ifft_x_step

        fr = np.interp(r, ifft_x, ifft_imag)
    else:
        raise NotImplementedError(
            "{} is not an allowed method for calculate_fr".format(method)
        )
    return Pattern(r, fr)


def calculate_sq_from_fr(fr_pattern: Pattern, q: np.ndarray) -> Pattern:
    """
    Calculates S(Q) from an F(r) pattern for given q values.

    :param fr_pattern:              input F(r) pattern
    :param q:                       numpy array giving the q-values for which S(q) will be calculated,

    :return: F(r) pattern
    """
    r, fr = fr_pattern.data
    sq = simpson(fr * np.array(np.sin(np.outer(r.T, q))).T, x=r) / q + 1

    # there should be a way of doing this with fft, but the current implementation is not working
    # correctly (it will not reproduce the S(q) pattern after transforming to fr and back accurately)
    # elif method == "fft":
    #     q_step = q[1] - q[0]
    #     r_step = r[1] - r[0]

    #     n_out = int(np.pi / (r_step * q_step))

    #     r_max_for_ifft = 2 * n_out * r_step
    #     ifft_x_step = 2 * np.pi / r_max_for_ifft
    #     ifft_x = np.arange(n_out) * ifft_x_step

    #     y_for_ifft = np.concatenate((fr, np.zeros(2 * n_out - len(r))))
    #     ifft_result = np.fft.ifft(y_for_ifft) * r_max_for_ifft
    #     ifft_imag = np.imag(ifft_result)[:n_out]

    #     sq = np.interp(q, ifft_x, ifft_imag) / q + 1
    return Pattern(q, sq)


def calculate_sq_from_gr(
    gr_pattern: Pattern,
    q: np.ndarray,
    atomic_density: float,
    method: str = "integral",
) -> Pattern:
    """
    Performs a back Fourier transform from the pair distribution function g(r)

    :param gr_pattern:      g(r) pattern
    :param q:               numpy array of q values for which S(Q) should be calculated
    :param atomic_density:  number_density in atoms/A^3

    :return: S(Q) pattern
    """
    r, gr = gr_pattern.data

    # removing the nan value at the first index, which is caused by the division by zero when r started from zero
    if np.isnan(gr[0]):
        gr[0] = 0
    fr_pattern = Pattern(r, (gr - 1) * (4.0 * np.pi * r * atomic_density))
    return calculate_sq_from_fr(fr_pattern, q)


def calculate_gr(fr_pattern: Pattern, atomic_density: float) -> Pattern:
    """
    Calculates a g(r) pattern from a given F(r) pattern and the atomic density

    :param fr_pattern:     F(r) pattern
    :param atomic_density:  atomic density in atoms/A^3

    :return: g(r) pattern
    """
    r, f_r = fr_pattern.data
    g_r = 1 + f_r / (4.0 * np.pi * r * atomic_density)
    return Pattern(r, g_r)



# -*- coding: utf-8 -*-

from copy import deepcopy

import numpy as np
from lmfit import Parameters, minimize

from . import Pattern
from .transform import calculate_fr, calculate_gr

__all__ = [
    "optimize_sq",
    "optimize_density",
]


def optimize_sq(
    sq_pattern: Pattern,
    r_cutoff: float,
    iterations: int,
    atomic_density: float,
    use_modification_fcn: bool = False,
    attenuation_factor: float = 1,
    fcn_callback=None,
    callback_period: int = 2,
    fourier_transform_method: str = "fft",
):
    """
    Performs an optimization of the structure factor based on an r_cutoff value as described in Eggert et al. 2002 PRB,
    65, 174105. This basically does back and forward transforms between S(Q) and f(r) until the region below the
    r_cutoff value is a flat line without any oscillations.

    :param sq_pattern:
        original S(Q)
    :param r_cutoff:
        cutoff value below which there is no signal expected (below the first peak in g(r))
    :param iterations:
        number of back and forward transforms
    :param atomic_density:
        density in atoms/A^3
    :param use_modification_fcn:
        Whether to use the Lorch modification function during the Fourier transform.
        Warning: When using the Lorch modification function, usually more iterations are needed to get to the
        wanted result.
    :param attenuation_factor:
        Sometimes the initial change during back and forward transformations results in a runaway, by setting the
        attenuation factor reduces the amount of change during each iteration.
    :param fcn_callback:
        Function which will be called at an iteration period defined by the callback_period parameter.
        The function should take three arguments: sq_pattern, fr_pattern and gr_pattern.
        Additionally, the function should return a boolean value, where True continues the optimization and False will
        stop the optimization.
    :param callback_period:
        determines how frequently the fcn_callback will be called.
    :param fourier_transform_method:
        determines which method will be used for the Fourier transform. Possible values are 'fft' and 'integral'

    :return:
        optimized S(Q) pattern
    """
    r = np.arange(0, r_cutoff, 0.02)
    sq_pattern = deepcopy(sq_pattern)
    for iteration in range(iterations):
        fr_pattern = calculate_fr(
            sq_pattern, r, use_modification_fcn, method=fourier_transform_method
        )
        q, sq_int = sq_pattern.data
        r, fr_int = fr_pattern.data

        delta_fr = fr_int + 4 * np.pi * r * atomic_density

        in_integral = np.array(np.sin(np.outer(q.T, r))) * delta_fr
        integral = np.trapz(in_integral, r) / attenuation_factor
        sq_optimized = sq_int * (1 - 1.0 / q * integral)

        sq_pattern = Pattern(q, sq_optimized)

        if fcn_callback is not None and iteration % callback_period == 0:
            fr_pattern = calculate_fr(
                sq_pattern,
                use_modification_fcn=use_modification_fcn,
                method=fourier_transform_method,
            )
            gr_pattern = calculate_gr(fr_pattern, atomic_density)
            fcn_callback(sq_pattern, fr_pattern, gr_pattern)
    return sq_pattern


from .calc import calculate_pdf
from .configuration import DataConfig, CalculationConfig
from .methods import ExtrapolationMethod


def optimize_density(
    data_config: DataConfig,
    calculation_config: CalculationConfig,
    type: str = "gr",
    min_range: tuple[float, float] = (0, 1),
    method: str = "lsq",
) -> tuple[float, float]:
    """
    Optimizes the density of the sample using the g(r) or S(Q) (chosen by the method parameter). The density in the
    Sample configuration of the CalculationConfig is taking as starting parameter

    For type='gr' the optimization is based on the g(r) function, and the density is optimized to minimize the
    low g(r) region to be close to zero. For better results, the g(r) function is calculated with the Lorch
    modification function. The general procedure is explained in Eggert et al. 2002 PRB, 65, 174105.

    For type='sq' the optimization is based on the low Q part of the S(Q) function, and the density is optimized
    to minimize the difference between the original S(Q) function without any optimization and the optimized S(Q)
    function. The configuration should have extrapolation enabled for this to work best.
    For polyatomic systems, finding the density using this procedure is much less susceptible to the Q_max value of
    the S(Q) than the g(r) based optimization. However, density is not exactly the same for both methods and the
    method needs to be verified further. (Please us the type='sq' with caution.)

    The best for both types is to have a reference density to compare it to. Based on this then further calculations
    of e.g. high pressure or high temperature densities can be performed.

    For this procedure to work best, the S(Q) optimization should be enabled in the calculation configuration. The
    chosen parameters are then used in the find density function.

    example usage:
    ```
    from glassure.calc import create_calculate_pdf_configs
    from glassure.optimization import optimize_density

    data_config, calculation_config = create_calculate_pdf_configs(data, composition, density, background)
    calculation_config.transform.q_min = 1
    calculation_config.transform.q_max = 16
    calculation_config.transform.extrapolation.method = ExtrapolationMethod.LINEAR
    calculation_config.optimize = OptimizeConfig(r_cutoff=1.4)

    density, error = optimize_density(data_config, calculation_config, type='gr', range=(0.1, 1.2))
    ```

    :param data_config:
        Data configuration
    :param calculation_config:
        Calculation configuration
    :param type:
        Method to use for the optimization. Possible values are 'gr' and 'sq'.
    :param min_range:
        x range of the data to use for the minimization to find the density. For method='gr' this is the r-range of the
        g(r) function to minimize to be close to zero. For method='sq' this is the Q-range of the S(Q) function to
        minimize the difference between the original and optimized S(Q) function.
    :param optimization_method:
        Method to use for the optimization. Possible values are 'nelder' and 'lsq'.

    :return:
        a tuple with two values:
        - the density and the standard error for optimization_method='lsq'
        - the density and the residual for optimization_method='nelder'
    """

    params = Parameters()
    params.add("density", value=calculation_config.sample.density, min=0.0, max=100)

    optim_config = calculation_config.model_copy(deep=True)
    optim_config.transform.use_modification_fcn = True

    if type == "sq":
        reference_config = calculation_config.model_copy(deep=True)
        reference_config.optimize = None
        reference_result = calculate_pdf(data_config, reference_config)

    def fcn(params):
        density = params["density"].value
        optim_config.sample.density = density
        result = calculate_pdf(data_config, optim_config)

        if type == "gr":
            r, gr = result.gr.limit(*min_range).data
            residual = np.trapz(gr**2, r)
        elif type == "sq":
            q, sq = result.sq.limit(*min_range).data
            sq_ref = reference_result.sq.limit(*min_range).y
            residual = np.trapz((sq - sq_ref) ** 2, q)
        return residual

    if method == "nelder":
        res = minimize(
            fcn,
            params,
            method="nelder",
            options={"maxfev": 500, "fatol": 0.0001, "xatol": 0.0001},
        )
        return res.params["density"].value, res.residual[0]
    elif method == "lsq":
        res = minimize(
            fcn,
            params,
            method="least_squares",
        )
        return res.params["density"].value, res.params["density"].stderr
    else:
        raise ValueError(f"Invalid optimization method: {method}")



# -*- coding: utf-8 -*-
import os
import re
from abc import ABC, abstractmethod

import numpy as np
import scipy
import pandas
from . import _module_path

module_data_path = os.path.join(_module_path(), 'data')

atomic_weights = pandas.read_csv(os.path.join(
    module_data_path, 'atomic_weights.csv'),
    index_col=0)


class ScatteringFactorCalculator(ABC):
    """
    Abstract class for scattering factor calculators.
    """

    @abstractmethod
    def get_coherent_scattering_factor(self, element: str, q):
        raise NotImplementedError

    @abstractmethod
    def get_incoherent_intensity(self, element: str, q):
        raise NotImplementedError

    @property
    def elements(self):
        raise NotImplementedError


class ScatteringFactorCalculatorHajdu(ScatteringFactorCalculator):
    """
    Scattering factor calculator based on the work of Hajdu et al. (Acta Cryst. (1992). A48, 344-352).
    """

    def __init__(self):
        self.coherent_param = pandas.read_csv(
            os.path.join(module_data_path, 'hajdu', 'param_coherent_scattering_factors.csv'),
            index_col=0)
        self.incoherent_param = pandas.read_csv(
            os.path.join(module_data_path, 'hajdu', 'param_incoherent_scattering_intensities.csv'),
            index_col=0)

    def get_coherent_scattering_factor(self, element: str, q):
        """
        Calculates the coherent scattering factor for a given element and q values.

        :param element: Element symbol
        :param q: q array
        :return: coherent scattering factor array
        """
        if element not in self.coherent_param.index.values:
            raise ElementNotImplementedException(element)
        fs_coh = 0
        s = q / (4 * np.pi)
        for ind in range(1, 5):
            A = self.coherent_param['A' + str(ind)][element]
            B = self.coherent_param['B' + str(ind)][element]
            fs_coh += A * np.exp(-B * s ** 2)

        C = self.coherent_param['C'][element]
        fs_coh += C
        return fs_coh

    def get_incoherent_intensity(self, element: str, q):
        """
        Calculates the incoherent scattering intensity for a given element and q values.

        :param element: Element symbol
        :param q: q array
        :return: incoherent scattering intensity array
        """
        fs_coherent = self.get_coherent_scattering_factor(element, q)
        intensity_coherent = fs_coherent ** 2
        s = q / (4 * np.pi)
        Z = float(self.incoherent_param['Z'][element])
        M = float(self.incoherent_param['M'][element])
        K = float(self.incoherent_param['K'][element])
        L = float(self.incoherent_param['L'][element])
        intensity_incoherent = (Z - intensity_coherent / Z) * (1 - M * (np.exp(-K * s) - np.exp(-L * s)))
        return intensity_incoherent

    @property
    def elements(self):
        """
        Returns a list of available elements.
        """
        return self.coherent_param.index.values


class ScatteringFactorCalculatorBrownHubbell(ScatteringFactorCalculator):
    """
    Scattering factor calculator based on the work of Brown et al., 2006 and Hubbell et al., 1975.
    """

    def __init__(self):
        self.coherent_params = pandas.read_csv(
            os.path.join(module_data_path, 'brown_hubbell', 'param_coherent_scattering_factors.csv'),
            index_col=0)

        self.incoherent_intensities = pandas.read_csv(
            os.path.join(module_data_path, 'brown_hubbell', 'incoherent_scattering_intensities.csv'))

    def get_coherent_scattering_factor(self, element: str, q):
        """
        Calculates the coherent scattering factor for a given element and q values.

        :param element: Element symbol
        :param q: q array
        :return: coherent scattering factor array
        """
        if element not in self.coherent_params.index.values:
            raise ElementNotImplementedException(element)
        fs_coh = 0
        s = q / (4 * np.pi)
        for ind in range(1, 5):
            A = self.coherent_params['a' + str(ind)][element]
            B = self.coherent_params['b' + str(ind)][element]
            fs_coh += A * np.exp(-B * s ** 2)

        C = self.coherent_params['c'][element]
        fs_coh += C
        return fs_coh

    def get_incoherent_intensity(self, element: str, q):
        """
        Calculates the incoherent scattering intensity for a given element and q values.

        :param element: Element symbol
        :param q: q array
        :return: incoherent scattering intensity array
        """
        # use regular expression to find element string of input
        element = re.findall('[A-zA-Z]*', element)[0]
        if element not in self.incoherent_intensities.keys():
            raise ElementNotImplementedException(element)
        interp = scipy.interpolate.interp1d(self.incoherent_intensities['q'], self.incoherent_intensities[element],
                                            kind='cubic')
        return interp(q)

    @property
    def elements(self):
        """
        Returns a list of available elements.
        """
        return self.coherent_params.index.values


calculators = {
    'hajdu': ScatteringFactorCalculatorHajdu(),
    'brown_hubbell': ScatteringFactorCalculatorBrownHubbell()
}

sources = calculators.keys()


def get_calculator(source: str) -> ScatteringFactorCalculator:
    """
    Returns the calculator for a given source. Possible sources are 'hajdu' and 'brown_hubbell'.
    """
    if source not in calculators.keys():
        raise SourceNotImplementedException(source)
    return calculators[source]


def get_available_elements(source: str) -> list[str]:
    """
    Returns a list of available elements for a given source. Possible sources are 'hajdu' and 'brown_hubbell'.
    """
    return get_calculator(source).elements


def calculate_coherent_scattering_factor(element: str, q: np.array, source: str = 'hajdu') -> np.array:
    """
    Calculates the coherent scattering factor for a given element and q values.

    :param element: Element symbol
    :param q: q array in A^-1
    :param source: Source of the scattering factors. Possible sources are 'hajdu' and 'brown_hubbell'.
    :return: coherent scattering factor array
    """
    return get_calculator(source).get_coherent_scattering_factor(element, q)


def calculate_incoherent_scattered_intensity(element: str, q: np.array, source: str = 'hajdu') -> np.array:
    """
    Calculates the incoherent scattering intensity for a given element and q values.

    :param element: Element symbol
    :param q: q array
    :param source: Source of the scattering factors. Possible sources are 'hajdu' and 'brown_hubbell'.
    :return: incoherent scattering intensity array
    """
    return get_calculator(source).get_incoherent_intensity(element, q)


class ElementNotImplementedException(Exception):
    def __init__(self, element):
        self.element = element

    def __str__(self):
        return repr('Element ' + self.element + ' not known or available.')


class SourceNotImplementedException(Exception):
    def __init__(self, source):
        self.source = source

    def __str__(self):
        return repr('Source ' + self.source + ' not known or available.')
